import logging
import argparse
import torch
import torch.nn as nn
from torchvision import models
from torchvision import datasets
from torchvision.transforms import ToTensor
from multiprocessing import Process
from pyconlai import ConLArguments, ConLPoCArguments, DSgd, FedDatasetsClassification

formatter = '%(asctime)s [%(name)s] %(levelname)s :  %(message)s'
logging.basicConfig(level=logging.INFO, format=formatter)

batch_size = 100
num_rounds = 100

def run_client(args, client_id, train_data_loader, test_data_loader, device="cpu"):
    round_idx = 0

    model = models.resnet18()

    # Preliminary: Optimizer can be anything
    org_optimizer = torch.optim.SGD(filter(lambda p: p.requires_grad, model.parameters()), lr=0.01)
    optimizer = DSgd(args, org_optimizer, model.parameters())

    criterion = nn.CrossEntropyLoss()

    logger = logging.getLogger("ConLAi-Tra%02d" % client_id)
    logger.info("Training Start!!")

    model = model.to(device)

    while round_idx < num_rounds:
        round_idx += 1
        model.train()

        batch_loss = []
        for batch_idx, (x, labels) in enumerate(train_data_loader):
            x, labels = x.to(device), labels.to(device)
            optimizer.zero_grad()
            labels = labels.long()

            log_probs = model(x)
            loss = criterion(log_probs, labels)
            loss.backward()
            batch_loss.append(loss.item())

            optimizer.step()

        model.eval()
        class_correct = list(0. for _ in range(10))
        class_total = list(0. for _ in range(10))
        criterion = nn.CrossEntropyLoss().to(device)

        metrics = {"accuracy": 0., "loss": 0.}
        loss_ary = []
        with torch.no_grad():
            total = 0
            correct = 0
            batch_loss = []
            for batch_idx, (x, target) in enumerate(test_data_loader):
                x = x.to(device)
                target = target.to(device)
                pred = model(x)
                target = target.long()
                loss = criterion(pred, target)

                _, predicted = torch.max(pred, 1)
                c = (predicted == target).squeeze()
                for i in range(4):
                    label = target[i]
                    class_correct[label] += c[i].item()
                    class_total[label] += 1
                total += target.size(0)
                correct += (predicted == target).sum().item()

                batch_loss.append(loss.item())
                loss_ary.append(sum(batch_loss) / len(batch_loss))

            metrics["accuracy"] = correct / total
            metrics["loss"] = sum(loss_ary) / len(loss_ary)

        logger.info("Round= {:4d} \tAcc: {:.6f} \tLoss: {:.6f} \tDiff: {:.8f} \tK: {:2d}".format(
            round_idx, metrics["accuracy"], metrics["loss"], optimizer.diff, optimizer.inner_loop))

        # round update
        optimizer.round_update(metrics)

    # close
    logger.info("Training Finished!!")



def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("config_path", type=str, help="path of config file")
    args = arg_parser.parse_args()

    net_args = ConLArguments.from_yml(args.config_path)
    poc_args = ConLPoCArguments.from_yml(args.config_path)

    train_data = datasets.CIFAR10(root=poc_args.data_cache_dir, train=True, download=True, transform=ToTensor())
    valid_data = datasets.CIFAR10(root=poc_args.data_cache_dir, train=False, download=True, transform=ToTensor())
    # Split the CIFAR10 dataset for each client
    fed_datasets = FedDatasetsClassification(net_args, poc_args, batch_size,
                                             train_data, valid_data, class_num=10)

    clients = []
    for client_id in range(poc_args.worker_num):
        client = Process(target=run_client, args=(net_args, client_id,
                                                  fed_datasets.fed_dataset(client_id)["train"],
                                                  fed_datasets.fed_dataset(client_id)["valid"],
                                                  "cuda"))
        client.start()
        clients.append(client)

    for client in clients:
        client.join()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from omegaconf import OmegaConf


@dataclass
class ConLArguments:
    server_url: str = "localhost:9200"
    repo_name: str = "conlai"
    ws_max_size: int = 1048576000

    # training args
    inner_loop: int = 10
    eta: float = 0.1  # Not used in DSGD

    @classmethod
    def from_yml(cls, yml_path: str):
        loaded = OmegaConf.load(yml_path)
        read_conf = ConLArguments(**loaded.net)

        base_conf = OmegaConf.structured(ConLArguments)
        merged = OmegaConf.merge(base_conf, read_conf)
        return cls(**merged)


@dataclass
class ConLPoCArguments:
    # data args
    data_cache_dir: str = "./data/"
    partition_method: str = "hetero"
    partition_alpha: float = 10.
    random_seed: int = 42

    # device args
    worker_num: int = 1

    @classmethod
    def from_yml(cls, yml_path: str):
        loaded = OmegaConf.load(yml_path)
        read_conf = ConLPoCArguments(**loaded.poc)

        base_conf = OmegaConf.structured(ConLPoCArguments)
        merged = OmegaConf.merge(base_conf, read_conf)
        return cls(**merged)



import numpy as np
from typing import Dict
from torch.utils.data import Dataset, DataLoader
from .mixin import FedDatasetsMixin, FedInnerLoopSampler
from ..context import ConLArguments, ConLPoCArguments


class FedDatasetsClassification(FedDatasetsMixin):
    def __init__(self, net_args: ConLArguments, poc_args: ConLPoCArguments, batch_size: int,
                 train_data: Dataset, valid_data: Dataset, class_num: int, min_len=10):
        super().__init__(net_args, poc_args, batch_size, train_data, valid_data, class_num)

        indices = self._partition_data(self._train_data, self._train_data_num, min_len)
        self._fed_train_data_loader, self._fed_train_data_num = self._build_datasets(self._train_data, indices)

        indices = self._partition_data(self._valid_data, self._valid_data_num, min_len)
        self._fed_valid_data_loader, _ = self._build_datasets(self._valid_data, indices)

    def _partition_data(self, dataset: Dataset, n_data: int, min_len: int):
        net_data_idx_map = {}
        np.random.seed(self._random_seed)
        target = np.array(dataset.targets)

        if self._partition_method == "hetero":
            min_size = 0
            while min_size < min_len:
                idx_batch = [[] for _ in range(self._clients_num)]
                # for each class in the dataset
                for k in range(self._class_num):
                    idx_k = np.where(target == k)[0]
                    np.random.shuffle(idx_k)
                    proportions = np.random.dirichlet(np.repeat(self._partition_alpha, self._clients_num))
                    # Balance
                    proportions = np.array(
                        [
                            p * (len(idx_j) < n_data / self._clients_num)
                            for p, idx_j in zip(proportions, idx_batch)
                        ]
                    )
                    proportions = proportions / proportions.sum()
                    proportions = (np.cumsum(proportions) * len(idx_k)).astype(int)[:-1]
                    idx_batch = [
                        idx_j + idx.tolist()
                        for idx_j, idx in zip(idx_batch, np.split(idx_k, proportions))
                    ]
                    min_size = min([len(idx_j) for idx_j in idx_batch])

            for j in range(self._clients_num):
                np.random.shuffle(idx_batch[j])
                net_data_idx_map[j] = idx_batch[j]
                self._logger.info("partition data hetero alpha= %.1f  CL=%d: datasize= %d / %d" %
                                  (self._partition_alpha, j, len(net_data_idx_map[j]), n_data))

        else:
            # partition_method = homo
            total_num = n_data
            indices = np.random.permutation(total_num)
            batch_indices = np.array_split(indices, self._clients_num)
            net_data_idx_map = {i: batch_indices[i].tolist() for i in range(self._clients_num)}
            for i in range(self._clients_num):
                self._logger.info("partition data homo CL=%d: datasize= %d / %d" %
                                  (i, len(net_data_idx_map[i]), n_data))
        return net_data_idx_map

    def _build_datasets(self, dataset: Dataset, indices: Dict) -> (Dict, Dict):
        dataloader = {}
        datasize = {}

        for client_idx in range(self._clients_num):
            _sampler = FedInnerLoopSampler(self._batch_size, self._inner_loop, indices[client_idx])
            dataloader[client_idx] = DataLoader(dataset=dataset, batch_size=self._batch_size, sampler=_sampler)
            datasize[client_idx] = len(indices[client_idx])

        return dataloader, datasize



from .mixin import ConLOptimizer


class DSgd(ConLOptimizer):
    STATE_P = "state_p"

    def __init__(self, args, org_optimizer, parameters):
        super(DSgd, self).__init__(args, org_optimizer, parameters)

    def __setstate__(self, state):
        super(DSgd, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    def aggregate(self, rcv_p):
        i = 0
        diff_sum = 0.
        snd_p = []
        state = None
        if rcv_p is not None:
            state = rcv_p[self.STATE_P]

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                if state is not None:
                    p_d = state[i].to(p.get_device())
                    diff_sum += self._criterion(p_d, p.data)
                    p.data = (p_d + p.data) / 2
                    i += 1
                snd_p.append(p.data.clone().to("cpu"))

        self._diff_latest = diff_sum
        return {self.STATE_P: snd_p}



import torch
import pickle
import websocket
from typing import Dict
from torch import nn
from torch.optim.optimizer import Optimizer
from abc import ABCMeta, abstractmethod
from ..pb import ConLParams, ConLMetrics


class ConLOptimizer(Optimizer, metaclass=ABCMeta):
    def __init__(self, args, org_optimizer, parameters):
        self._org_optimizer = org_optimizer
        self._server_url = args.server_url
        self._round = 0
        self._inner_loop = args.inner_loop
        self._inner_cnt = 0
        self._diff_latest = 0.
        self._criterion = nn.MSELoss(reduction="sum")

        ws_target_url = "ws://%s/ws" % args.server_url
        self._ws = websocket.create_connection(ws_target_url)

        defaults = dict(eta=args.eta)
        super().__init__(filter(lambda p: p.requires_grad, parameters), defaults)

    def __del__(self):
        self._ws.close()

    @torch.no_grad()
    def precondition(self):
        pass

    @abstractmethod
    def aggregate(self, rcv_p: Dict) -> Dict:
        pass

    @torch.no_grad()
    def step(self, closure=None):
        self.precondition()

        loss = self._org_optimizer.step(closure)

        self._inner_cnt += 1
        if self._inner_cnt >= self.inner_loop:
            # pull
            msg = ConLParams()
            msg.op = "pull"
            self._ws.send(msg.SerializeToString(), websocket.ABNF.OPCODE_BINARY)
            message = self._ws.recv()
            msg.ParseFromString(message)

            params = None
            if len(msg.params) > 0:
                params = pickle.loads(msg.params)
            merged = self.aggregate(params)

            # push
            msg.op = "push"
            msg.params = pickle.dumps(merged)
            self._ws.send(msg.SerializeToString(), websocket.ABNF.OPCODE_BINARY)
            self._inner_cnt = 0

        return loss

    def round_update(self, metrics: Dict):
        self._round += 1
        metrics_list = []
        for k, v in metrics.items():
            mt = ConLMetrics()
            mt.name = k
            mt.value = v
            metrics_list.append(mt)

        msg = ConLParams()
        msg.op = "update"
        msg.stats.round = self._round
        msg.stats.metrics.extend(metrics_list)

        self._ws.send(msg.SerializeToString(), websocket.ABNF.OPCODE_BINARY)
        self._ws.recv()

    @property
    def optimizer(self):
        return self._org_optimizer

    @property
    def lr(self):
        lr = 0.
        for group in self._org_optimizer.param_groups:
            lr = group['lr']
        return lr

    @property
    def inner_loop(self):
        return self._inner_loop

    @property
    def diff(self):
        return self._diff_latest




from typing import List
from logging import getLogger
from abc import ABCMeta
from torch.utils.data import Dataset, DataLoader, sampler
from ..context import ConLArguments, ConLPoCArguments


class FedDatasetsMixin(metaclass=ABCMeta):
    def __init__(self, net_args: ConLArguments, poc_args: ConLPoCArguments,
                 batch_size: int, train_data: Dataset, valid_data: Dataset, class_num=0):
        self._train_data = train_data
        self._valid_data = valid_data
        self._class_num = class_num
        self._batch_size = batch_size
        self._inner_loop = net_args.inner_loop
        self._clients_num = poc_args.worker_num
        self._random_seed = poc_args.random_seed

        self._partition_method = poc_args.partition_method
        self._partition_alpha = poc_args.partition_alpha

        self._train_data_loader = DataLoader(dataset=self._train_data, batch_size=self._batch_size,
                                             shuffle=True, drop_last=True)
        self._valid_data_loader = DataLoader(dataset=self._valid_data, batch_size=self._batch_size,
                                             shuffle=False, drop_last=True)
        self._train_data_num = len(self._train_data)
        self._valid_data_num = len(self._valid_data)

        self._fed_train_data_num = {}
        self._fed_train_data_loader = {}
        self._fed_valid_data_loader = {}

        self._logger = getLogger("FedDatasets")

    def fed_dataset(self, client_id: int):
        return {
            "train": self._fed_train_data_loader[client_id],
            "valid": self._fed_valid_data_loader[client_id],
            "num": self._fed_train_data_num[client_id]
        }

    @property
    def train_data_loader(self) -> DataLoader:
        return self._train_data_loader

    @property
    def valid_data_loader(self) -> DataLoader:
        return self._valid_data_loader

    @property
    def class_num(self):
        return self._class_num


class FedInnerLoopSampler(sampler.Sampler[int]):
    def __init__(self, batch_size: int, inner_loop: int, indices: List):
        super().__init__()
        self._n_data = len(indices)
        self._batch_size = batch_size
        self._inner_loop = inner_loop

        self._indices = []
        self._data_indices = indices
        self._n_data_batch = self._batch_size * self._inner_loop if self._inner_loop is not None else self._n_data
        self._n_offset = 0

    def __len__(self):
        return self._n_data_batch

    def __iter__(self):
        # Removal of used data
        self._indices = self._indices[self._n_offset:]
        self._n_offset = 0  # reset start index

        # Prepare at least 1 epoch's worth of data in the index list that stores the data call order.
        while len(self._indices) <= self._n_data_batch:
            self._indices += self._data_indices

        sidx, eidx = self._n_offset, self._n_offset + self._n_data_batch
        indices = self._indices[sidx:eidx].copy()
        self._n_offset = eidx
        yield from indices



from .context import ConLArguments, ConLPoCArguments
from .optimizer import DSgd
from .datasets import FedDatasetsClassification

__all__ = [
    "ConLArguments",
    "ConLPoCArguments",
    "DSgd",
    "FedDatasetsClassification",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: conl.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'conl.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nconl.proto\"*\n\x0b\x43onLMetrics\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02\"F\n\tConLStats\x12\x0b\n\x03src\x18\x01 \x01(\t\x12\r\n\x05round\x18\x02 \x01(\x05\x12\x1d\n\x07metrics\x18\x03 \x03(\x0b\x32\x0c.ConLMetrics\"C\n\nConLParams\x12\n\n\x02op\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x01(\x0c\x12\x19\n\x05stats\x18\x04 \x01(\x0b\x32\n.ConLStatsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'conl_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CONLMETRICS']._serialized_start=14
  _globals['_CONLMETRICS']._serialized_end=56
  _globals['_CONLSTATS']._serialized_start=58
  _globals['_CONLSTATS']._serialized_end=128
  _globals['_CONLPARAMS']._serialized_start=130
  _globals['_CONLPARAMS']._serialized_end=197
# @@protoc_insertion_point(module_scope)
